import os
import shutil
from unittest import TestCase
from pathlib import Path

from foliant.contrib.utils import prepend_file


def rel_name(path: str):
    return os.path.join(os.path.dirname(__file__), path)


TEST_DATA_PATH = Path(os.path.dirname(__file__)) / 'test_data' / 'utils'


class TestPrependFile(TestCase):
    def setUp(self):
        self.content = \
'''Prepended paragraph

Second prepended paragraph\n\n'''

        for source in TEST_DATA_PATH.glob('*.md'):
            if not str(source).startswith('exp'):
                cp = source.parent / f'cp_{source.name}'
                shutil.copyfile(source, cp)

    def tearDown(self):
        for source in TEST_DATA_PATH.glob('cp_*.md'):
            os.remove(source)

    def test_simple(self):
        filepath = TEST_DATA_PATH / 'cp_simple.md'
        with open(TEST_DATA_PATH / 'exp_simple.md') as f:
            expected = f.read()

        prepend_file(filepath, self.content)

        with open(TEST_DATA_PATH / 'cp_simple.md') as f:
            result = f.read()

        self.assertEqual(result, expected)

    def test_heading_before(self):
        filepath = TEST_DATA_PATH / 'cp_heading.md'
        with open(TEST_DATA_PATH / 'exp_heading_before.md') as f:
            expected = f.read()

        prepend_file(filepath, self.content, before_heading=True)

        with open(TEST_DATA_PATH / 'cp_heading.md') as f:
            result = f.read()

        self.assertEqual(result, expected)

    def test_heading_not_before(self):
        filepath = TEST_DATA_PATH / 'cp_heading.md'
        with open(TEST_DATA_PATH / 'exp_heading_not_before.md') as f:
            expected = f.read()

        prepend_file(filepath, self.content, before_heading=False)

        with open(TEST_DATA_PATH / 'cp_heading.md') as f:
            result = f.read()

        self.assertEqual(result, expected)

    def test_yfm_before(self):
        filepath = TEST_DATA_PATH / 'cp_yfm.md'
        with open(TEST_DATA_PATH / 'exp_yfm_before.md') as f:
            expected = f.read()

        prepend_file(filepath, self.content, before_yfm=True)

        with open(TEST_DATA_PATH / 'cp_yfm.md') as f:
            result = f.read()

        self.assertEqual(result, expected)

    def test_yfm_not_before(self):
        filepath = TEST_DATA_PATH / 'cp_yfm.md'
        with open(TEST_DATA_PATH / 'exp_yfm_not_before.md') as f:
            expected = f.read()

        prepend_file(filepath, self.content, before_yfm=False)

        with open(TEST_DATA_PATH / 'cp_yfm.md') as f:
            result = f.read()

        self.assertEqual(result, expected)



from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import Iterator
from typing import Union


def flatten_seq(seq: Union[list, dict]) -> list:
    """convert a sequence of embedded sequences into a plain list"""

    result = []
    vals = seq.values() if isinstance(seq, dict) else seq
    for i in vals:
        if isinstance(i, (dict, list)):
            result.extend(flatten_seq(i))
        elif isinstance(i, str):
            result.append(i)
    return result


class ChapterNotFoundError(Exception):
    pass


class Chapters:
    """
    Helper class converting chapter list of complicated structure
    into a plain list of chapter names or path to actual md files
    in the src dir.
    """

    def __init__(self,
                 chapters: list,
                 working_dir: Optional[Union[str, Path]] = None,
                 src_dir: Optional[Union[str, Path]] = None,
                 ):
        self.working_dir = Path(working_dir).resolve() if working_dir else None
        self.src_dir = Path(src_dir).resolve() if src_dir else None
        self._chapters = chapters
        self._flat = flatten_seq(chapters)

    def __len__(self) -> int:
        return len(self._flat)

    def __getitem__(self, ind: int) -> str:
        return self._flat[ind]

    def __contains__(self, item: str) -> bool:
        return item in self._flat

    def __iter__(self):
        return iter(self._flat)

    def __repr__(self) -> str:
        return f'Chapters({self._chapters})'

    @classmethod
    def from_config(cls, config: dict) -> 'Chapters':
        '''Returns a Chapters instance, initiated with properties from config.'''
        return cls(
            config['chapters'],
            working_dir=config['tmp_dir'],
            src_dir=config['src_dir']
        )

    @property
    def chapters(self) -> list:
        """Original chapters list"""
        return self._chapters

    @chapters.setter
    def chapters(self, chapters) -> None:
        self._chapters = chapters
        self._flat = flatten_seq(chapters)

    @property
    def flat(self) -> list:
        """Flat list of chapter file names"""
        return self._flat

    def get_chapter_by_path(self, filepath: Union[str, Path]) -> str:
        """
        Try and find filepath in working dir or src dir. If it is present in
        one of those — return relative path to file (as it is stated in chapters)
        """
        abs_path = Path(filepath).resolve()
        if self.working_dir and self.working_dir in abs_path.parents:
            chapter_path = str(abs_path.relative_to(self.working_dir))
        if self.src_dir and self.src_dir in abs_path.parents:
            chapter_path = str(abs_path.relative_to(self.src_dir))
        if chapter_path and chapter_path in self.chapters:
            return chapter_path
        else:
            raise ChapterNotFoundError(f'{filepath} is not in the chapter list')

    def paths(self, parent_dir: Union[str, Path]) -> Iterator:
        """
        Returns generator yielding PosixPath object with chapter path, relative
        to parent_dir.
        """

        return (Path(parent_dir) / chap for chap in self.flat)

    def get_chapter_title(self, chapter_path: str) -> str:
        """
        Returns the title for the chapter, defined in the chapters list.
        Returns empty string if chapter defined without title.
        Raises ChapterNotFoundError if chapter is not present in the list.

        Examples:

        ***
        chapters:
            - Chapter Title: info/index.md

        get_chapter_title('info/index.md') will return 'Chapter Title'

        ***
        chapters:
            - info/index.md

        get_chapter_title('info/index.md') will return ''

        ***
        chapters:
            - info.md

        get_chapter_title('info/index.md') will raise ChapterNotFoundError

        :param chapter_path: path to the chapter for which the title needs to be
                             found, _as it is stated in the chapters list_.

        :returns: chapter title or empty string.
        """
        def find_chapter(chapters: Union[list, dict], to_find: str) -> Union[str, dict, None]:
            if isinstance(chapters, list):
                for i in chapters:
                    if isinstance(i, (list, dict)):
                        found = find_chapter(i, to_find)
                        if found:
                            return found
                    elif isinstance(i, str) and i == to_find:
                        return i
            elif isinstance(chapters, dict):
                val = next(iter(chapters.values()))
                if isinstance(val, str) and val == to_find:
                    return chapters
                else:
                    found = find_chapter(val, to_find)
                    if found:
                        return found
            elif isinstance(chapters, str):
                if chapters == to_find:
                    return chapters
            return None

        chapter = find_chapter(self.chapters, chapter_path)
        if chapter is None:
            raise ChapterNotFoundError(f'{chapter_path} is not in the chapter list')
        if isinstance(chapter, str):
            return ''
        elif isinstance(chapter, dict):
            return next(iter(chapter.keys()))



from concurrent.futures import ThreadPoolExecutor
from os import makedirs, listdir
from os.path import join
from pathlib import Path
from shutil import move
from typing import Union


def prepend_file(
    filepath: Union[str, Path],
    content: str,
    before_yfm: bool = False,
    before_heading: bool = True
):
    '''
    Insert `content` at the beginning of the file `filepath`.

    :param filepath: path to file which needs to be prepended.
    :param content: content to be inserted.
    :before_yfm: if file starts with YAML Front Matter, insert content before it
    :before_heading: if file starts with a heading, insert content before it
    '''
    with open(filepath, encoding='utf8') as f:
        source = f.read()

    start = 0

    if not before_yfm and source.startswith('---\n'):
        yfm_end = source.find('\n---\n', 1)
        start = yfm_end + len('\n---\n') if yfm_end != -1 else 0
        # add line break for not to break the heading
        content = '\n' + content
    if not before_heading and source.startswith('#'):
        start = source.find('\n', 1) + 1
        if start == 0:
            start = len(source)
        # add line break for not to break the heading
        content = '\n' + content

    processed_content = source[:start] + content + source[start:]

    with open(filepath, 'w', encoding='utf8') as f:
        f.write(processed_content)


def move_files_threadpool(src: Path, dest: Path, n_workers: int = 64):
    """Move files from one directory to another with ThreadPoolExecutor

    :param src: path to source directory
    :param dest: path to the destination directory
    :param n_workers: number of workers in ThreadPool
    """
    def move_files(src_paths, dest_dir):
        for src_path in src_paths:
            move(src_path, dest_dir)

    makedirs(dest, exist_ok=True)
    files = [join(src, name) for name in listdir(src)]
    chunksize = round(len(files) / n_workers)
    chunksize = 1 if chunksize == 0 else chunksize
    with ThreadPoolExecutor(n_workers) as exe:
        for i in range(0, len(files), chunksize):
            filenames = files[i:(i + chunksize)]
            _ = exe.submit(move_files, filenames, dest)



import os
import re
import traceback
import threading

from pathlib import Path
from typing import Callable, Optional, Union
from functools import wraps

from foliant.preprocessors.base import BasePreprocessor
from foliant.utils import output


MAX_THREADS = max(1, os.cpu_count() - 2)
thread_semaphore = threading.Semaphore(MAX_THREADS)

def run_in_thread(enabled=True):
    """
    If enabled=False, returns the original function immediately without wrapping.
    """
    def actual_decorator(fn):
        if not enabled:
            return fn

        @wraps(fn)
        def wrapper(*args, **kwargs):
            def thread_task():
                try:
                    fn(*args, **kwargs)
                finally:
                    thread_semaphore.release()

            thread_semaphore.acquire()
            thread = threading.Thread(target=thread_task, daemon=True)
            thread.start()
            return thread

        return wrapper
    
    return actual_decorator

def allow_fail(msg: str = 'Failed to process tag. Skipping.') -> Callable:
    """
    If function func fails for some reason, warning is issued but preprocessor
    doesn't terminate. In this case the tag remains unchanged.
    Decorator issues a warning to user with BasePreprocessorExt _warning method.
    If first positional argument is a match object, it is passed as context.
    """

    def decorator(func):
        def wrapper(self, *args, **kwargs) -> Optional[str]:
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                if args and isinstance(args[0], re.Match):
                    self._warning(f'{msg} {e}',
                                  context=self.get_tag_context(args[0]),
                                  error=e)
                    return args[0].group(0)
                else:
                    self._warning(f'{msg} {e}', error=e)
                    return None
        return wrapper
    return decorator


class BasePreprocessorExt(BasePreprocessor):
    """Extension of BasePreprocessor with useful helper methods"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.current_filename = ''
        self.current_pos = 0
        self.current_func = None
        self.buffer = {}

    @staticmethod
    def get_tag_context(match: re.Match,
                        limit: int = 100,
                        full_tag: bool = False) -> str:
        '''
        Get context of the tag match object.

        Returns a string with <limit> symbols before match, the match string and
        <limit> symbols after match.

        If full_tag == False, matched string is limited too: first <limit>/2
        symbols of match and last <limit>/2 symbols of match.
        '''

        source = match.string
        start = max(0, match.start() - limit)  # index of context start
        end = min(len(source), match.end() + limit)  # index of context end
        span = match.span()  # indeces of match (start, end)
        result = '...' if start != 0 else ''  # add ... at beginning if cropped
        if span[1] - span[0] > limit and not full_tag:  # if tag contents longer than limit
            bp1 = match.start() + limit // 2
            bp2 = match.end() - limit // 2
            result += f'{source[start:bp1]} <...> {source[bp2:end]}'
        else:
            result += source[start:end]
        if end != len(source):  # add ... at the end if cropped
            result += '...'
        return result

    def _warning(self,
                 msg: str,
                 context: str = '',
                 error: Exception = None,
                 debug_msg: str = '') -> None:
        '''
        Log warning and print to user.

        If debug mode — print also context (if sepcified) and error (if specified).

        :param msg:       — message which should be logged;
        :param context:   — tag context got with get_tag_context function. If
                            specified — will be logged. If debug = True it
                            will also go to STDOUT.
        :param error:     — exception which was caught before warning. If specified —
                            error traceback whill be added to log (and debug output) message.
        :param debug_msg: — message to additionally print to stdout in debug mode.
        '''

        output_message = ''
        if self.current_filename:
            output_message += f'[{self.current_filename}] '
        output_message += msg + '\n'
        log_message = output_message
        if debug_msg:
            log_message += f'{debug_msg}\n'
        if context:
            log_message += f'Context:\n---\n{context}\n---\n'
        if error:
            tb_str = traceback.format_exception(etype=type(error),
                                                value=error,
                                                tb=error.__traceback__)
            log_message += '\n'.join(tb_str)
        if self.debug:
            output_message = log_message
        output(f'WARNING: {output_message}', self.quiet)
        self.logger.warning(log_message)

    def pos_injector(self, block: re.Match) -> str:
        """
        Save offset of match object to self.current_pos and run
        self.current_func with this match object.
        """

        self.current_pos = block.start()
        return self.current_func(block)

    def save_file(self, path: Union[str, Path], content: str) -> None:
        with open(path, 'w', encoding='utf8') as f:
            f.write(content)

    def _process_tags_for_all_files(self,
                                    func: Callable,
                                    log_msg: str = 'Applying preprocessor',
                                    buffer: bool = False) -> None:
        '''
        Apply function func to all Markdown-files in the working dir

        :param func: function that should be applied to each found tag. Function
                     must accept 1 parameter: regex match object (found tag)
        :param log_msg: message text which will be logged at the beginning
        :param buffer: if True, processed text of each file will be buffered and
                       at the end all files will be saved at once.
        '''
        self.logger.info(log_msg)

        multithread = self.context['config'].get('multithread', False)
        @run_in_thread(enabled=multithread)
        def process(self, markdown_file_path):
            self.current_filepath = Path(markdown_file_path)
            self.current_filename = str(self.current_filepath.
                                        relative_to(self.working_dir))

            with open(markdown_file_path,
                      encoding='utf8') as markdown_file:
                content = markdown_file.read()

            self.current_func = func
            self.current_pos = 0
            processed_content = self.pattern.sub(self.pos_injector, content)

            if isinstance(processed_content, str):
                if buffer:
                    self.buffer[markdown_file_path] = processed_content
                else:
                    self.save_file(markdown_file_path, processed_content)
        for markdown_file_path in self.working_dir.rglob('*.md'):
            process(self, markdown_file_path)
        self.current_filename = ''

        for path, content in self.buffer.items():
            self.save_file(path, content)
        self.buffer = {}

    def _process_all_files(self,
                           func: Callable,
                           log_msg: str = 'Applying preprocessor',
                           buffer: bool = False) -> None:
        '''Apply function func to all Markdown-files in the working dir'''
        self.logger.info(log_msg)

        multithread = self.context['config'].get('multithread', False)
        @run_in_thread(enabled=multithread)
        def process(markdown_file_path):
            self.current_filepath = Path(markdown_file_path)
            self.current_filename = str(self.current_filepath.
                                        relative_to(self.working_dir))

            with open(markdown_file_path,
                      encoding='utf8') as markdown_file:
                content = markdown_file.read()

            processed_content = func(content)
            if isinstance(processed_content, str):
                if buffer:
                    self.buffer[markdown_file_path] = processed_content
                else:
                    self.save_file(markdown_file_path, processed_content)
        for markdown_file_path in self.working_dir.rglob('*.md'):
            process(markdown_file_path)
        self.current_filename = ''

        for path, content in self.buffer.items():
            self.save_file(path, content)
        self.buffer = {}



import os
from unittest import TestCase
from pathlib import Path

from foliant.contrib.chapters import Chapters
from foliant.contrib.chapters import ChapterNotFoundError
from foliant.contrib.chapters import flatten_seq

from .utils import chcwd


class TestFlattenSeq(TestCase):
    def test_plain_list(self):
        seq = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md']
        self.assertEqual(flatten_seq(seq), seq)

    def test_nested_lists(self):
        seq = [
            'ch1.md',
            'ch2.md',
            [
                'ch3.md',
                'ch4.md'
            ],
            'ch5.md',
            [
                ['ch6.md', 'ch7.md'],
            ],
            [
                'ch8.md',
                [
                    ['ch9.md'],
                    [
                        'ch10.md',
                        ['ch11.md']
                    ],
                    'ch12.md',
                ],
                'ch13.md'
            ]
        ]

        expected = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md', 'ch5.md', 'ch6.md', 'ch7.md', 'ch8.md', 'ch9.md', 'ch10.md', 'ch11.md', 'ch12.md', 'ch13.md']

        self.assertEqual(flatten_seq(seq), expected)

    def test_dict(self):
        seq = {'Title1': 'ch1.md', 'Title2': 'ch2.md', 'Title3': 'ch3.md'}
        expected = ['ch1.md', 'ch2.md', 'ch3.md']
        self.assertEqual(flatten_seq(seq), expected)

    def test_nested_lists_and_dicts(self):
        seq = [
            'ch1.md',
            {'Title2': 'ch2.md'},
            ['ch3.md', 'ch4.md'],
            [
                {
                    'Title5': 'ch5.md',
                    'Title6': [
                        'ch6.md',
                        'ch7.md'
                    ]
                }
            ],
            {'Title8': 'ch8.md'},
        ]

        expected = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md', 'ch5.md', 'ch6.md', 'ch7.md', 'ch8.md']

        self.assertEqual(flatten_seq(seq), expected)


class TestChapters(TestCase):
    test_data_path = 'test/test_data/chapters'

    def test_from_config(self):
        chapters = ['ch1.md', 'ch2.md', {'Title3': 'ch3.md'}, {'Title4': ['ch4.md', 'ch5.md']}]
        config = {'chapters': chapters, 'tmp_dir': '__folianttmp__', 'src_dir': 'src'}
        with chcwd(self.test_data_path):
            chapters_obj = Chapters.from_config(config)

        expected_flat = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md', 'ch5.md']
        expected_workingdir = (Path(self.test_data_path) / '__folianttmp__').resolve()
        expected_srcdir = (Path(self.test_data_path) / 'src').resolve()

        self.assertEqual(chapters_obj.chapters, chapters)
        self.assertEqual(chapters_obj.flat, expected_flat)
        self.assertEqual(chapters_obj.working_dir, expected_workingdir)
        self.assertEqual(chapters_obj.src_dir, expected_srcdir)

    def test_chapters_setter(self):
        chapters = ['ch1.md', 'ch2.md', {'Title3': 'ch3.md'}, {'Title4': ['ch4.md', 'ch5.md']}]
        expected_flat = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md', 'ch5.md']
        chapters_obj = Chapters({})

        chapters_obj.chapters = chapters

        self.assertEqual(chapters_obj.chapters, chapters)
        self.assertEqual(chapters_obj.flat, expected_flat)

    def test_get_chapter_by_path(self):
        chapters = ['ch1.md', 'ch2.md', 'ch3.md']
        with chcwd(self.test_data_path):
            chapters_obj = Chapters(chapters,
                                    working_dir='__folianttmp__',
                                    src_dir='src')
            ch1 = 'src/ch1.md'
            ch1_full = os.path.join(os.getcwd(), 'src/ch1.md')
            ch2_path = Path('src/ch2.md').resolve()
            ch3 = 'src/ch3.md'
            ch4 = 'src/ch4.md'

            self.assertEqual(chapters_obj.get_chapter_by_path(ch1), 'ch1.md')
            self.assertEqual(chapters_obj.get_chapter_by_path(ch1_full), 'ch1.md')
            self.assertEqual(chapters_obj.get_chapter_by_path(ch2_path), 'ch2.md')

            # present only in src:
            self.assertEqual(chapters_obj.get_chapter_by_path(ch3), 'ch3.md')

            with self.assertRaises(ChapterNotFoundError):
                chapters_obj.get_chapter_by_path(ch4)

    def test_paths(self):
        chapters = ['ch1.md', 'ch2.md', {'Title3': 'ch3.md'}, {'Title4': ['ch4.md', 'ch5.md']}]
        flat_chapters = ['ch1.md', 'ch2.md', 'ch3.md', 'ch4.md', 'ch5.md', ]
        with chcwd(self.test_data_path):
            src_path = Path('src').resolve()
            workdir_path = Path('__folianttmp__').resolve()
            src_paths = [src_path / p for p in flat_chapters]
            workdir_paths = [workdir_path / p for p in flat_chapters]

            chapters_obj = Chapters(chapters,
                                    working_dir='__folianttmp__',
                                    src_dir='src')

            self.assertEqual(
                list(chapters_obj.paths(chapters_obj.src_dir)),
                src_paths
            )
            self.assertEqual(
                list(chapters_obj.paths(chapters_obj.working_dir)),
                workdir_paths
            )

    def test_get_chapter_title(self):
        chapters = [
            'ch1.md',
            'ch2.md',
            {'Title3': 'ch3.md'},
            {'Title4': ['ch4.md', 'ch5.md']},
            {
                'Section Title': [
                    'ch6.md',
                    {'Title7': 'ch7.md'}
                ]
            }
        ]

        chapters_obj = Chapters(chapters)
        self.assertEqual(chapters_obj.get_chapter_title('ch3.md'), 'Title3')
        self.assertEqual(chapters_obj.get_chapter_title('ch6.md'), '')
        self.assertEqual(chapters_obj.get_chapter_title('ch7.md'), 'Title7')
        with self.assertRaises(ChapterNotFoundError):
            chapters_obj.get_chapter_title('nonexistant.md')



from unittest import TestCase

from foliant.contrib.header_anchors import make_unique_confluence
from foliant.contrib.header_anchors import make_unique_mkdocs
from foliant.contrib.header_anchors import make_unique_pandoc
from foliant.contrib.header_anchors import make_unique_slate
from foliant.contrib.header_anchors import to_id_aglio
from foliant.contrib.header_anchors import to_id_confluence
from foliant.contrib.header_anchors import to_id_mdtopdf
from foliant.contrib.header_anchors import to_id_pandoc
from foliant.contrib.header_anchors import to_id_slate


class TestToIdPandoc(TestCase):
    def test_strip_everything_before_first_letter(self):
        self.assertEqual(to_id_pandoc('qwerty'), 'qwerty')
        self.assertEqual(to_id_pandoc('___qwerty'), 'qwerty')
        self.assertEqual(to_id_pandoc('127qwerty'), 'qwerty')
        self.assertEqual(to_id_pandoc('$%^&*(qwerty'), 'qwerty')

    def test_accept_alphanumeric_underscore_hyphen_dot(self):
        self.assertEqual(to_id_pandoc('q1werty'), 'q1werty')
        self.assertEqual(to_id_pandoc('q12-werty'), 'q12-werty')
        self.assertEqual(to_id_pandoc('q12-w_erty'), 'q12-w_erty')
        self.assertEqual(to_id_pandoc('q12-w_e.rty'), 'q12-w_e.rty')

    def test_remove_unsupported(self):
        self.assertEqual(to_id_pandoc('qw%^&erty'), 'qwerty')
        self.assertEqual(to_id_pandoc('qw^er$t#y'), 'qwerty')

    def test_lower(self):
        self.assertEqual(to_id_pandoc('QwErTy'), 'qwerty')
        self.assertEqual(to_id_pandoc('QWERTY'), 'qwerty')

    def test_space_to_hyphen(self):
        self.assertEqual(to_id_pandoc('q w e r t y'), 'q-w-e-r-t-y')
        self.assertEqual(to_id_pandoc('qwe       rty'), 'qwe-rty')

    def test_all_removed(self):
        self.assertEqual(to_id_pandoc('&*#@'), 'section')


class TestMdToPdf(TestCase):
    def test_lower(self):
        self.assertEqual(to_id_mdtopdf('QwErTy'), 'qwerty')
        self.assertEqual(to_id_mdtopdf('QWERTY'), 'qwerty')

    def test_accept_alphanumeric_underscore_hyphen(self):
        self.assertEqual(to_id_mdtopdf('q1werty'), 'q1werty')
        self.assertEqual(to_id_mdtopdf('q12-werty'), 'q12-werty')
        self.assertEqual(to_id_mdtopdf('q12-w_erty'), 'q12-w_erty')

    def test_remove_unsupported(self):
        self.assertEqual(to_id_mdtopdf('qw%^&erty'), 'qwerty')
        self.assertEqual(to_id_mdtopdf('qw^er$t#y'), 'qwerty')

    def test_space_to_hyphen(self):
        self.assertEqual(to_id_mdtopdf('q w e r t y'), 'q-w-e-r-t-y')
        self.assertEqual(to_id_mdtopdf('qwe       rty'), 'qwe-------rty')


class TestAglio(TestCase):
    def test_lower(self):
        self.assertEqual(to_id_aglio('QwErTy'), 'header-qwerty')
        self.assertEqual(to_id_aglio('QWERTY'), 'header-qwerty')

    def test_special_symbols_to_hyphens(self):
        self.assertEqual(to_id_aglio('q w"e/r:t<y'), 'header-q-w-e-r-t-y')
        self.assertEqual(to_id_aglio('q=w>e\\rty'), 'header-q-w-e-rty')

    def test_dashes_and_commas_collapse(self):
        self.assertEqual(to_id_aglio('q-w-e-r-t-y'), 'header-q-w-e-r-t-y')
        self.assertEqual(to_id_aglio('q,w,e,r,t,y'), 'header-q,w,e,r,t,y')
        self.assertEqual(to_id_aglio('q,,w,,,e,,,,r,,,,,t,,,,,,y'), 'header-q,w,e,r,t,y')

    def test_beautify_replace(self):
        self.assertEqual(to_id_aglio("q'w...e---r--ty"), 'header-q’w…e—r–ty')


class TestToIdConfluence(TestCase):
    def test_remove_spaces(self):
        self.assertEqual(to_id_confluence('q w e r t y'), 'qwerty')
        self.assertEqual(to_id_confluence('qwe       rty'), 'qwerty')

    def test_beautify_replace(self):
        self.assertEqual(to_id_confluence("q'w...e---r--ty"), 'q’w…e—r–ty')


# uslugify tested here:
# https://github.com/facelessuser/pymdown-extensions/blob/main/tests/test_extensions/test_slugs.py

class TestToIdSlate(TestCase):
    def test_special_symbols_to_hyphens(self):
        self.assertEqual(to_id_slate('q w_e%%$r#()t+y17'), 'q-w_e-r-t-y17')

    def test_html_escape(self):
        self.assertEqual(to_id_slate('''q&w"e'r>t<y'''), 'q-amp-w-quot-e-39-r-gt-t-lt-y')

    def test_strip_sep(self):
        self.assertEqual(to_id_slate('%%%-++qwerty###@'), 'qwerty')

    def test_remove_tags(self):
        self.assertEqual(to_id_slate('<a href="asd">qwerty</a> very <b>BIG</b> change'), 'qwerty-very-big-change')

    def test_hash_for_empty_string(self):
        self.assertEqual(to_id_slate('%# -!!@'), 'bcb3049620')


class TestMakeUniqueMkdocs(TestCase):
    def test_first_occurence_no_change(self):
        self.assertEqual(make_unique_mkdocs('qwerty', 1), 'qwerty')

    def test_empty_string(self):
        self.assertEqual(make_unique_mkdocs('', 1), '_1')
        self.assertEqual(make_unique_mkdocs('', 7), '_7')

    def test_second_plus_occurence(self):
        self.assertEqual(make_unique_mkdocs('qwerty', 2), 'qwerty_1')
        self.assertEqual(make_unique_mkdocs('qwerty', 7), 'qwerty_6')


class TestMakeUniquePandoc(TestCase):
    def test_first_occurence_no_change(self):
        self.assertEqual(make_unique_pandoc('qwerty', 1), 'qwerty')

    def test_second_plus_occurence(self):
        self.assertEqual(make_unique_pandoc('qwerty', 2), 'qwerty-1')
        self.assertEqual(make_unique_pandoc('qwerty', 7), 'qwerty-6')


class TestMakeUniqueConfluence(TestCase):
    def test_first_occurence_no_change(self):
        self.assertEqual(make_unique_confluence('qwerty', 1), 'qwerty')

    def test_second_plus_occurence(self):
        self.assertEqual(make_unique_confluence('qwerty', 2), 'qwerty.1')
        self.assertEqual(make_unique_confluence('qwerty', 7), 'qwerty.6')


class TestMakeUniqueSlate(TestCase):
    def test_first_occurence_no_change(self):
        self.assertEqual(make_unique_slate('qwerty', 1), 'qwerty')

    def test_second_plus_occurence(self):
        self.assertEqual(make_unique_slate('qwerty', 2), 'qwerty-2')
        self.assertEqual(make_unique_slate('qwerty', 7), 'qwerty-7')



import yaml

from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Collection
from typing import Sequence
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

class ValidationError(Exception):
    '''Error for validations when validation is failed'''
    pass


class RequiredParamsMissingError(Exception):
    '''Error for required params checking when requirement is not satisfied'''
    pass


class Options:
    '''
    Helper class which imitates dictionary with options but has some
    handy methods like option validation and conversion.
    '''

    def __init__(self,
                 options: dict,
                 defaults: Optional[Dict[str, Any]] = None,
                 convertors: Optional[Dict[str, Callable]] = None,
                 validators: Optional[Dict[str, Callable]] = None,
                 required: Optional[Union[List[str], List[List[str]]]] = None):
        '''
        options (dict)    — options dictionary,
        defaults (dict)   — dictionary with default values,
        convertors (dict) — dictionary with key = option name, value = function
                            which will be applied to the value of this option
                            before storing in class.
        validators (dict) — dictionary with key = option name, value = function
                            which will be applied to the value of this option.
                            Function should check for validity and raise
                            ValidationError if the check fails.
        required (list)   - list of required params or list of tuples with
                            combinations of required params.
        '''
        self.defaults = dict(defaults) if defaults else {}
        self._options = {**self.defaults, **options}
        self._validators = dict(validators) if validators else {}
        self._convertors = dict(convertors) if convertors else {}
        self._required = list(required) if required else []
        self.validate()
        self._convert()

    @property
    def options(self) -> dict:
        '''Actual options dictionary'''
        return self._options

    def validate(self) -> None:
        '''
        Validate all options with supplied validators and check for required
        params.
        Raises ValidationError if any of validation checks fails.
        Raises RequiredParamsMissingError if required params are not supplied.
        '''
        def _check_required(combination: Collection) -> bool:
            for param in combination:
                if param not in self.options:
                    return False
            return True

        for key in self._validators:
            if key in self.options:
                try:
                    self._validators[key](self.options[key])
                except ValidationError as e:
                    raise ValidationError(f'Error in option "{key}": {e}')
        if self._required:
            if isinstance(self._required[0], str):
                if not _check_required(self._required):
                    raise RequiredParamsMissingError(
                        f'Not all required params are supplied: {self._required}')
            else:  # several combinations of required params are possible
                check_result = any(_check_required(comb) for comb in self._required)
                if not check_result:
                    required_combs = "\nor:\n".\
                        join(str(comb).strip('()[]') for comb in self._required)
                    raise RequiredParamsMissingError(
                        f'Not all required params are supplied. '
                        f'Required parameter combinations are:\n{required_combs}')

    def _convert(self) -> None:
        '''
        Convert all options with supplied convertors and replace values in
        options dictionary in place.
        '''
        if not self._convertors:
            return

        for key in self._convertors:
            if key in self.options:
                convertor = self._convertors[key]
                self.options[key] = convertor(self.options[key])

    def is_default(self, option: str) -> bool:
        '''return True if option value is same as default'''
        if option in self.defaults:
            return self.options[option] == self.defaults[option]
        return False

    def __str__(self):
        return f'<{self.__class__.__name__}{self.options}>'

    def __getitem__(self, ind: str) -> Any:
        return self.options[ind]

    def __setitem__(self, ind: str, val: Any):
        self.options[ind] = val
        self.validate()

    def __contains__(self, ind: str) -> bool:
        return ind in self.options

    def __iter__(self):
        return iter(self.options.keys())

    def __bool__(self) -> bool:
        return bool(self.options)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self.options.get(key, default)

    def keys(self):
        return self.options.keys()

    def items(self):
        return self.options.items()

    def values(self):
        return self.options.values()


class CombinedOptions(Options):
    '''
    Helper class which combines several Options objects into one. If options
    interlap the one to be returned is defined by 'priority' or by the order
    defined.

    Apart from that it is a normal Options object.
    '''

    def __init__(self,
                 options: dict,
                 priority: Optional[Union[str, Sequence[str]]] = None,
                 defaults: Optional[Dict[str, Any]] = None,
                 convertors: Optional[Dict[str, Callable]] = None,
                 validators: Optional[Dict[str, Callable]] = None,
                 required: Optional[Union[List[str], List[List[str]]]] = None):
        '''
        :param options:  dictionary where key = priority,
                         value = option dictionary.
        :param priority: initial priority  or list of prioritites.

        other parameters are same as in parent
        '''
        self.defaults = dict(defaults) if defaults else {}
        self._options_dict = dict(options) if options else {}
        self._validators = dict(validators) if validators else {}
        self._convertors = dict(convertors) if convertors else {}
        self._required = list(required) if required else []

        if isinstance(priority, (list, tuple)):
            self.priority = list(priority)
        else:
            self.priority = priority  # type: ignore

    @property
    def priority(self) -> Union[str, Sequence[str]]:
        '''returns current priority'''
        return self._priority

    @priority.setter
    def priority(self, val: Union[str, Sequence[str], None]) -> None:
        '''sets new priority and updates active options dictionary'''
        if isinstance(val, (list, tuple)):
            priority_list = val
        elif isinstance(val, str):
            priority_list = [val]
        else:
            priority_list = []

        for p in priority_list:
            if p not in self._options_dict:
                raise ValueError('Priority must be one of: '
                                 f'{", ".join(self._options_dict.keys())}. '
                                 f'Value received: {p}')
        self._priority = priority_list
        self.set_options()

    def set_options(self) -> None:
        '''
        Sets new active options dict with options combined from all options
        dicts with priority according to self.priority.
        '''

        self._options = deepcopy(self.defaults)

        for key in reversed(list(self._options_dict)):
            if key not in self.priority:
                self._options.update(self._options_dict[key])

        for priority in reversed(self.priority):
            self._options.update(self._options_dict[priority])

        self.validate()
        self._convert()


def validate_in(supported: Collection, msg: Optional[str] = None) -> Callable:
    '''
    Simple validator factory. Resulting function checks if option value
    is contained in supported collection.

    `supported` may be any collection-like object with __contains__ method.
    Raises ValueError otherwise.

    msg is message given to the ValiadtionError.

    Returns a validator function.'''

    DEFAULT_MSG = 'Unsupported option value {val}. Should be one '\
                  'of: {supported}'

    def validate(val: Any) -> None:
        if val not in supported:
            raise ValidationError(message.format(val=val, supported=', '.join(str(s) for s in supported)))

    if not hasattr(supported, '__contains__'):
        raise ValueError('`supported` should be a collection')

    message = msg if msg else DEFAULT_MSG

    return validate


NoneType = type(None)


def val_type(supported: Union[List[Union[type, NoneType]], type]) -> Callable:
    '''
    Validator factory for validating param type.

    `supported` may be a type to check, None or a list (tuple) of possible types.
    '''

    MSG = 'Unsupported option value {val}. Must be of type {supported}'

    def validate(val: Any) -> None:
        for type_ in types:
            if type_ is None:
                if val is None:
                    return
            elif isinstance(val, type_):
                return
        raise ValidationError(MSG.format(val=val, supported=', '.join(str(t) for t in types)))
    if supported is None:
        types = [None]
    elif isinstance(supported, type):
        types = [supported]
    elif isinstance(supported, Collection):
        types = supported  # type: ignore
    else:
        raise ValueError('`supported` should be a type, None or a collection of types')
    return validate


def validate_exists(val: Union[str, Path]) -> None:
    '''Validator that checks if path specified in val exists'''
    MSG = 'Path {val} does not exist.'
    if val and not Path(val).exists():
        raise ValidationError(MSG.format(val=val))


def path_convertor(option: Union[str, Path]) -> Path:
    '''convert string to Path'''
    if isinstance(option, str):
        return Path(yaml.load(option, yaml.Loader))
    elif isinstance(option, Path):
        return option


def yaml_to_dict_convertor(option):
    '''
    DEPRECATED. Foliant does it automatically since 1.0.9
    convert yaml string or dict to dict
    '''

    if type(option) is dict:
        return option
    elif type(option) is str:
        return yaml.load(option, yaml.Loader)


def boolean_convertor(option: Any) -> bool:
    '''
    Convert option to bool if necessary.

    Accepts True\False, 'tRuE' \ 'falSE', 1\0, Y \ n, yes \ no

    Other types are validated as bool(object)
    '''
    str_dict = {
        '1': True,
        '0': False,
        'y': True,
        'n': False,
        'yes': True,
        'no': False,
        'true': True,
        'false': False
    }
    if type(option) == bool:
        return option
    elif type(option) == str:
        return str_dict.get(option.lower().strip(), True)
    else:
        return bool(option)


def rel_path_convertor(parent_path: Union[str, Path]) -> Callable:
    '''
    Convertor factory which makes option path relative to parent_path supplied
    during the convertor initialization.
    '''

    def _convertor(option: Union[str, Path]) -> Any:
        if not option:
            return option
        else:
            return Path(parent_path) / option

    return _convertor



import re
import unicodedata

from hashlib import sha1
from urllib.parse import quote
from typing import Dict

FALLBACK_BACKEND = 'pandoc'
FLAT_BACKENDS = ['pandoc', 'slate', 'aglio', 'mdtopdf']


class IDGenerator:
    '''
    This class generates IDs but counts the number of times they were called.
    Before returning the generated ID it passes it through the make_unique
    function to generate the unique ID.
    '''

    def __init__(self, backend: str):
        self.registry: Dict[str, int] = {}
        self.backend = backend

    def generate(self, heading: str) -> str:
        '''
        Generate the unique ID for the `heading`.

        :param heading: heading to be converted to ID.

        :returns: unique ID for the heading.
        '''
        id_ = to_id(heading, self.backend)
        self.registry[id_] = self.registry.setdefault(id_, 0) + 1
        return make_unique(id_, self.registry[id_], self.backend)

    def reset(self) -> None:
        '''
        Reset all ID count to 0.
        '''
        self.registry = {}


def is_flat(backend: str) -> bool:
    '''Determine whether backend is flat or not.'''
    return backend in FLAT_BACKENDS


def to_id(input_: str, backend: str) -> str:
    '''
    Convert a heading title to proper id for backend. If backend not found,
    fallback backend is used.

    :param input_: heading title
    :param backend: name of the backend

    :returns: converted id
    '''
    BACKEND_MAP = {
        'pandoc': to_id_pandoc,
        'mdtopdf': to_id_mdtopdf,
        'aglio': to_id_aglio,
        'mkdocs': to_id_mkdocs,
        'slate': to_id_slate,
        'confluence': to_id_confluence,
        'no_transform': to_id_no_transform,
    }
    if backend not in BACKEND_MAP:
        backend = FALLBACK_BACKEND
    return BACKEND_MAP[backend](input_)


def make_unique(input_: str, occurrence: int, backend: str) -> str:
    '''
    Make an id, converted from the title, unique according to backend rules.
    If backend not found, fallback backend is used.

    :param input_: potentially dublicate header id
    :param occurrence: number of occurrence of the header with this title
    :param backend: name of the backend

    :returns: unique id
    '''
    BACKEND_MAP = {
        'pandoc': make_unique_pandoc,
        'mdtopdf': make_unique_pandoc,
        'aglio': make_unique_pandoc,
        'mkdocs': make_unique_mkdocs,
        'slate': make_unique_slate,
        'confluence': make_unique_confluence,
    }
    if backend not in BACKEND_MAP:
        backend = FALLBACK_BACKEND
    return BACKEND_MAP[backend](input_, occurrence)


def to_id_pandoc(input_: str) -> str:
    """
    Quote from docs:

    The default algorithm used to derive the identifier from the heading text is:

        * Remove all formatting, links, etc.
        * Remove all footnotes.
        * Remove all non-alphanumeric characters, except underscores, hyphens, and periods.
        * Replace all spaces and newlines with hyphens.
        * Convert all alphabetic characters to lowercase.
        * Remove everything up to the first letter (identifiers may not begin with a number or punctuation mark).
        * If nothing is left after this, use the identifier section.
    """
    def accept(char: str) -> bool:
        if char in ALPHA:
            return True
        elif char.isalpha():
            return True
        elif char.isdigit():
            return True
        return False
    ALPHA = '_-.'
    result = ''
    source = input_.lower()
    accum = False

    # strip everything before first letter
    while source and not source[0].isalpha():
        source = source[1:]

    for char in source:
        if accept(char):
            if accum:
                accum = False
                result += f'-{char.lower()}'
            else:
                result += char.lower()
        elif char.isspace():
            accum = True
        else:
            pass

    if not result:
        return 'section'
    else:
        return result


def to_id_mdtopdf(input_: str) -> str:
    def accept(char: str) -> bool:
        return char in ALPHA or char.isalpha() or char.isdigit()
    ALPHA = '_-'
    result = ''
    source = input_.lower().strip()

    result = re.sub(r'\s', '-', result)

    for char in source:
        if accept(char):
            result += char
        elif char.isspace():
            result += '-'
        else:  # remove all other symbols
            pass

    return result


def to_id_aglio(input_: str) -> str:
    repl_dict = {
        "'": '’',
        '...': '…',
        '---': '—',
        '--': '–',
    }
    result = 'header-' + input_.lower().strip()

    for source, repl in repl_dict.items():
        result = result.replace(source, repl)
    result = re.sub(r'[\s"/:<=>\\]', '-', result)
    # result = re.sub(r'-+', '-', result)
    result = re.sub(r',+', ',', result)
    return result


def to_id_confluence(input_: str) -> str:
    repl_dict = {
        "'": '’',
        '...': '…',
        '---': '—',
        '--': '–',
    }
    result = re.sub(r'\s', '', input_)

    for source, repl in repl_dict.items():
        result = result.replace(source, repl)
    return result


def slugify(value: str, separator: str) -> str:
    """
    Slugify a string, to make it URL friendly.

    This function is copied from `markdown` python package
    (https://github.com/Python-Markdown/markdown/).

    Mkdocs uses it by default to generate heading IDs
    """
    normalized = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore')
    shaved = re.sub(r'[^\w\s-]', '', normalized.decode('ascii')).strip().lower()
    return re.sub(r'[%s\s]+' % separator, separator, shaved)


RE_TAGS = re.compile(r'</?[^>]*>', re.UNICODE)
NO_CASED = 0
UNICODE_CASED = 1
RE_ASCII_LETTERS = re.compile(r'[A-Z]', re.UNICODE)
RE_INVALID_SLUG_CHAR = re.compile(r'[^\w\- ]', re.UNICODE)
RE_SEP = re.compile(r' ', re.UNICODE)


def uslugify(text: str,
             sep:str,
             cased: int = NO_CASED,
             percent_encode: bool = False):
    """
    Unicode slugify (`utf-8`).

    This function is copied from `pymdownx.slugs` python package
    (https://github.com/facelessuser/pymdown-extensions/).
    """

    # Normalize, Strip html tags, strip leading and trailing whitespace, and lower
    slug = RE_TAGS.sub('', unicodedata.normalize('NFC', text)).strip()

    if cased == NO_CASED:
        slug = slug.lower()
    elif cased == UNICODE_CASED:

        def lower(m):
            """Lowercase character."""
            return m.group(0).lower()

        slug = RE_ASCII_LETTERS.sub(lower, slug)

    # Remove non word characters, non spaces, and non dashes, and convert spaces to dashes.
    slug = RE_SEP.sub(sep, RE_INVALID_SLUG_CHAR.sub('', slug))

    return quote(slug.encode('utf-8')) if percent_encode else slug


def to_id_mkdocs(input_: str) -> str:
    return uslugify(input_, '-')


def parameterize_slate(string_to_clean: str, sep: str = '-') -> str:
    """
    Port of Rails `parameterize` function by Vinicius Quaiato.
    """
    HTML_ESCAPE = {
        "&": "&amp;",
        '"': "&quot;",
        "'": "&39;",
        "<": "&lt;",
        ">": "&gt;"
    }

    parameterized_string = unicodedata.normalize('NFKD', string_to_clean).encode('ASCII', 'ignore').decode()

    for char, repl in HTML_ESCAPE.items():
        parameterized_string = parameterized_string.replace(char, repl)

    parameterized_string = re.sub("[^a-zA-Z0-9_]+", sep, parameterized_string)

    if sep is not None and sep != '':
        parameterized_string = parameterized_string.strip(sep)
    return parameterized_string.lower()


def to_id_slate(input_: str) -> str:
    # removing tags
    source = re.sub(r'<[/\w_:=" ]*>', '', input_)
    source = parameterize_slate(source)
    return source if source else sha1(input_.encode('utf-8')).hexdigest()[:10]


def to_id_no_transform(input_: str) -> str:
    return input_


def make_unique_mkdocs(input_: str, occurrence: int = 1) -> str:
    if input_ == '':
        occurrence += 1
    if occurrence == 1:
        return input_
    else:
        return input_ + '_' + str(occurrence - 1)


def make_unique_pandoc(input_: str, occurrence: int = 1) -> str:
    if occurrence == 1:
        return input_
    else:
        return input_ + '-' + str(occurrence - 1)


def make_unique_confluence(input_: str, occurrence: int = 1) -> str:
    if occurrence == 1:
        return input_
    else:
        return input_ + '.' + str(occurrence - 1)


def make_unique_slate(input_: str, occurrence: int = 1) -> str:
    if occurrence == 1:
        return input_
    else:
        return input_ + f'-{occurrence}'



from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock

from foliant.contrib.combined_options import CombinedOptions
from foliant.contrib.combined_options import Options
from foliant.contrib.combined_options import RequiredParamsMissingError
from foliant.contrib.combined_options import ValidationError
from foliant.contrib.combined_options import boolean_convertor
from foliant.contrib.combined_options import path_convertor
from foliant.contrib.combined_options import rel_path_convertor
from foliant.contrib.combined_options import val_type
from foliant.contrib.combined_options import validate_exists
from foliant.contrib.combined_options import validate_in


class TestOptions(TestCase):
    def test_no_processing(self):
        original = {'key': 'val', 'int': 12, 'bool': True}
        options = Options(original)

        self.assertEqual(options.options, original)

    def test_defaults(self):
        original = {'key': 'val', 'int': 12, 'bool': True, 'overridden': 42}
        defaults = {'defaultkey': 'defaultvalue', 'overridden': 0}

        expected = {'key': 'val', 'int': 12, 'bool': True, 'overridden': 42, 'defaultkey': 'defaultvalue'}
        options = Options(original, defaults=defaults)
        self.assertEqual(options.defaults, defaults)
        self.assertEqual(options.options, expected)

    def test_required_flat(self):
        required = ['req1', 'req2']
        set1 = {'key': 'val', 'req1': 1, 'req2': False}

        options = Options(set1, required=required)

        set2 = {'key': 'val', 'req2': False}
        with self.assertRaises(RequiredParamsMissingError):
            options = Options(set2, required=required)

    def test_required_combinations(self):
        required = [['req1', 'req2'], ['req21', 'req22'], ['req']]
        set1 = {'key': 'val', 'req1': 1, 'req2': False}

        options = Options(set1, required=required)

        set2 = {'key': 'val', 'req2': False}
        with self.assertRaises(RequiredParamsMissingError):
            options = Options(set2, required=required)

        set3 = {'key': 'val', 'req21': False, 'req22': ''}
        options = Options(set3, required=required)

        set4 = {'key': 'val', 'req': '1'}
        options = Options(set4, required=required)

    def test_validate(self):
        mock_validator1 = Mock(return_value=None)
        mock_validator2 = Mock(return_value=None)
        original = {'key': 'val', 'int': 12, 'bool': True}
        options = Options(
            original,
            validators={'key': mock_validator1, 'bool': mock_validator2}
        )

        self.assertEqual(options['key'], 'val')
        self.assertEqual(options['int'], 12)
        self.assertEqual(options['bool'], True)

        mock_validator1.assert_called_once_with('val')
        mock_validator2.assert_called_once_with(True)

    def test_validation_error(self):
        mock_validator = Mock(side_effect=[ValidationError])
        original = {'key': 'val', 'int': 12, 'bool': True}
        with self.assertRaises(ValidationError):
            options = Options(
                original,
                validators={'int': mock_validator}
            )

    def test_convert(self):
        mock_convertor = Mock(return_value='converted')
        original = {'key': 'val', 'int': 12, 'bool': True}
        options = Options(
            original,
            convertors={'int': mock_convertor}
        )

        self.assertEqual(options['int'], 'converted')
        mock_convertor.assert_called_once_with(12)

    def test_is_default(self):
        original = {'key': 'val', 'def1': 12, 'def3': 'overridden'}
        defaults = {'def1': 12, 'def2': 'Default', 'def3': 'Default'}

        options = Options(original, defaults=defaults)
        self.assertTrue(options.is_default('def1'))
        self.assertTrue(options.is_default('def2'))
        self.assertFalse(options.is_default('def3'))
        self.assertFalse(options.is_default('key'))

class TestCombinedOptions(TestCase):
    def test_combine(self):
        options1 = {'key1': 'val1', 'key2': 'val2'}
        options2 = {'key3': 'val3', 'key4': 'val4'}
        coptions = CombinedOptions({'o1': options1, 'o2': options2})

        expected = {**options1, **options2}
        self.assertEqual(coptions.options, expected)

    def test_combine_override(self):
        options1 = {'key1': 'val1', 'key2': 'val2'}
        options2 = {'key2': 'val21', 'key4': 'val4'}
        coptions = CombinedOptions({'o1': options1, 'o2': options2})

        expected = {**options2, **options1}
        self.assertEqual(coptions.options, expected)

    def test_single_priority(self):
        options1 = {'key1': 'val1', 'key2': 'val2'}
        options2 = {'key2': 'val21', 'key4': 'val4'}
        coptions = CombinedOptions(
            {'o1': options1, 'o2': options2},
            priority='o2'
        )

        expected = {**options1, **options2}
        self.assertEqual(coptions.options, expected)

    def test_priority_list(self):
        options1 = {'key1': 'val1', 'key2': 'val2', 'key3': 'val3'}
        options2 = {'key2': 'val22', 'key3': 'val32', 'key4': 'val42'}
        options3 = {'key2': 'val23', 'key4': 'val43'}
        expected = {'key1': 'val1', 'key2': 'val23', 'key3': 'val32', 'key4': 'val43'}
        coptions = CombinedOptions(
            {'o1': options1, 'o2': options2, 'o3': options3},
            priority=['o3', 'o2', 'o1']
        )

        self.assertEqual(coptions.options, expected)

        coptions = CombinedOptions(
            {'o1': options1, 'o2': options2, 'o3': options3},
            priority=['o3', 'o2']
        )

        self.assertEqual(coptions.options, expected)

    def test_set_priority(self):
        options1 = {'key1': 'val1', 'key2': 'val2', 'key3': 'val3'}
        options2 = {'key2': 'val22', 'key3': 'val32', 'key4': 'val42'}
        options3 = {'key2': 'val23', 'key4': 'val43'}
        expected1 = {'key1': 'val1', 'key2': 'val23', 'key3': 'val32', 'key4': 'val43'}
        coptions = CombinedOptions(
            {'o1': options1, 'o2': options2, 'o3': options3},
            priority=['o3', 'o2', 'o1']
        )

        self.assertEqual(coptions.options, expected1)

        coptions.priority = ['o2', 'o3', 'o1']
        expected2 = {'key1': 'val1', 'key2': 'val22', 'key3': 'val32', 'key4': 'val42'}

        self.assertEqual(coptions.options, expected2)

        coptions.priority = 'o1'
        expected2 = {'key1': 'val1', 'key2': 'val2', 'key3': 'val3', 'key4': 'val42'}

        self.assertEqual(coptions.options, expected2)


class TestValidateIn(TestCase):
    def test_validation_pass(self):
        vals = [1, 3]
        supported = [1, 2, 3]
        validator = validate_in(supported)
        for val in vals:
            validator(val)

    def test_validation_fail(self):
        val = 'Unsupported'
        supported = [1, 2, 3]
        validator = validate_in(supported)
        with self.assertRaises(ValidationError):
            validator(val)

    def test_custom_error_message(self):
        val = 'Unsupported'
        supported = [1, 2, 3]
        validator = validate_in(supported, msg='Custom message')
        with self.assertRaises(ValidationError) as caught:
            validator(val)

        self.assertEqual(str(caught.exception), 'Custom message')


class TestValType(TestCase):
    def test_validation_pass_single(self):
        val = 'string'
        supported = str
        validator = val_type(supported)
        validator(val)

    def test_validation_pass_multiple(self):
        vals = ['string', 12, True]
        supported = [str, int]
        validator = val_type(supported)
        for val in vals:
            validator(val)

    def test_validation_pass_none(self):
        val = None
        supported = None
        validator = val_type(supported)
        validator(val)

    def test_validation_fail_single(self):
        val = 1
        supported = str
        validator = val_type(supported)
        with self.assertRaises(ValidationError):
            validator(val)

    def test_validation_fail_multiple(self):
        val = 1.12
        supported = (str, int)
        validator = val_type(supported)
        with self.assertRaises(ValidationError):
            validator(val)


class TestValidateExists(TestCase):
    def test_validation_pass_str(self):
        val = 'README.md'
        validate_exists(val)

    def test_validation_pass_path(self):
        val = Path('README.md')
        validate_exists(val)

    def test_validation_fail(self):
        val = Path('wrong.exe')
        with self.assertRaises(ValidationError):
            validate_exists(val)


class TestPathConvertor(TestCase):
    def test_str(self):
        converted = path_convertor('file.txt')
        self.assertEqual(converted, Path('file.txt'))

    def test_path(self):
        converted = path_convertor(Path('file.txt'))
        self.assertEqual(converted, Path('file.txt'))


class TestBooleanConvertor(TestCase):
    def test_true(self):
        true_vals = ['1', 'y', 'yes', 'true', True]
        for val in true_vals:
            self.assertTrue(boolean_convertor(val))

    def test_false(self):
        false_vals = ['0', 'n', 'no', 'false', False]
        for val in false_vals:
            self.assertFalse(boolean_convertor(val))

    def test_other(self):
        self.assertFalse(boolean_convertor(0))
        self.assertFalse(boolean_convertor(0.0))
        self.assertFalse(boolean_convertor([]))
        self.assertTrue(boolean_convertor(12))
        self.assertTrue(boolean_convertor(12.12))
        self.assertTrue(boolean_convertor(Path()))
        self.assertTrue(boolean_convertor('random string'))


class TestRelPathConvertor(TestCase):
    def test_str(self):
        parent = 'parent/path'
        convertor = rel_path_convertor(parent)

        self.assertEqual(convertor('rel/path'), Path('parent/path/rel/path'))

    def test_path(self):
        parent = Path('parent/path')
        convertor = rel_path_convertor(parent)

        self.assertEqual(convertor('rel/path'), Path('parent/path/rel/path'))
